import cluster from 'node:cluster';
import { format as dateFormat } from 'date-fns';
import config from '@/config/index.js';
import { envOption, LOG_LEVELS } from '@/env.js';

export type Level = LOG_LEVELS[keyof LOG_LEVELS];

/**
 * Class that facilitates recording log messages to the console.
 */
export default class Logger {
	private domain: string;
	private parentLogger: Logger | null = null;
	private store: boolean;
	/**
	 * Messages below this level will be discarded.
	 */
	private minLevel: Level;

	/**
	 * Create a logger instance.
	 * @param domain Logging domain
	 * @param store Whether to store messages
	 */
	constructor(domain: string, store = true, minLevel?: Level) {
		this.domain = domain;
		this.store = store;
		this.minLevel = minLevel ?? envOption.logLevel;
	}

	/**
	 * Create a child logger instance.
	 * @param domain Logging domain
	 * @param store Whether to store messages
	 * @returns A Logger instance whose parent logger is this instance.
	 */
	public createSubLogger(domain: string, store = true, minLevel?: Level): Logger {
		const logger = new Logger(domain, store, minLevel);
		logger.parentLogger = this;
		return logger;
	}

	/**
	 * Log a message.
	 * @param level Indicates the level of this particular message. If it is
	 *     less than the minimum level configured, the message will be discarded.
	 * @param message The message to be logged.
	 * @param subDomains Names of sub-loggers to be added.
	 */
	private log(level: Level, message: string, subDomains: Domain[] = [], _store = true): void {
		const store = _store && this.store;

		// Check against the configured log level.
		if (level < this.minLevel) return;

		// If this logger has a parent logger, delegate the actual logging to it,
		// so the parent domain(s) will be logged properly.
		if (this.parentLogger) {
			this.parentLogger.log(level, message, [this.domain].concat(subDomains), store);
			return;
		}

		const time = dateFormat(new Date(), 'HH:mm:ss');
		const worker = cluster.isPrimary ? '*' : cluster.worker?.id;
		const domains = [this.domain].concat(subDomains);

		let levelDisplay;
		switch (level) {
			case LOG_LEVELS.error:
				levelDisplay = 'ERR ';
				break;
			case LOG_LEVELS.warning:
				levelDisplay = 'WARN';
				break;
			case LOG_LEVELS.success:
				levelDisplay = 'DONE';
				break;
			case LOG_LEVELS.info:
				levelDisplay = 'INFO';
				break;
			case LOG_LEVELS.debug:
			default:
				levelDisplay = 'VERB';
				break;
		}

		let log = `${levelDisplay} ${worker}\t[${domains.join(' ')}]\t${message}`;
		if (envOption.withLogTime) log = time + ' ' + log;

		console.log(log);
	}

	/**
	 * Log an error message.
	 * Use in situations where execution cannot be continued.
	 * @param err Error or string containing an error message
	 */
	public error(err: string | Error): void {
		if (err instanceof Error) {
			this.log(LOG_LEVELS.error, err.toString());
		} else if (typeof err === 'object') {
			this.log(LOG_LEVELS.error, `${(err as any).message || (err as any).name || err}`);
		} else {
			this.log(LOG_LEVELS.error, `${err}`);
		}
	}

	/**
	 * Log a warning message.
	 * Use in situations where execution can continue but needs to be improved.
	 * @param message Warning message
	 */
	public warn(message: string): void {
		this.log(LOG_LEVELS.warning, message);
	}

	/**
	 * Log a success message.
	 * Use in situations where something has been successfully done.
	 * @param message Success message
	 */
	public succ(message: string): void {
		this.log(LOG_LEVELS.success, message);
	}

	/**
	 * Log a debug message.
	 * Use for debugging (information needed by developers but not required by users).
	 * @param message Debug message
	 */
	public debug(message: string): void {
		this.log(LOG_LEVELS.debug, message);
	}

	/**
	 * Log an informational message.
	 * Use when something needs to be logged but doesn't fit into other levels.
	 * @param message Info message
	 */
	public info(message: string): void {
		this.log(LOG_LEVELS.info, message);
	}
}
